"""
Statistics required for RCMIP analysis
"""
import numpy as np
import pandas as pd
import scipy.optimize
from scipy.stats import multivariate_normal, norm


def _diff_exp_sigma(sigma, ratio, conflevel):

    low = (1 - conflevel) / 2
    upp = 1 - low
    qno = norm.ppf(upp)
    ratio_sigma = (np.exp(qno * sigma) - 1) / (1 - np.exp(-qno * sigma))

    diff = ratio_sigma - ratio

    return diff**2


def _get_skewed_normal_width(ratio, conflevel):
    ratio_in = ratio if ratio >= 1 else 1 / ratio
    if not 0 <= conflevel <= 1:
        raise ValueError(f"Confidence level must be in [0, 1]. Received: {conflevel}")

    opt_res = scipy.optimize.minimize(
        _diff_exp_sigma,
        (0.1,),
        args=(ratio_in, conflevel),
        bounds=((10**-7, 10),),
        options={"maxiter": 10**6},
    )

    if not opt_res.success:
        raise ValueError(
            "Optimisation failed for ratio {} and conflevel {}".format(ratio, conflevel)
        )

    skewed_normal_width = float(opt_res.x.squeeze())

    return skewed_normal_width


def get_skewed_normal(median, lower, upper, conf, input_data):
    """
    Get skewed normal distribution matching the inputs

    Parameters
    ----------
    median : float
        Median of the output distribution

    lower : float
        Lower bound of the confidence interval

    upper : float
        Upper bound of the confidence interval

    conf : float
        Confidence associated with the interval [lower, upper] e.g. 0.66
        would mean that [lower, upper] defines the 66% confidence range

    input_data : :obj:`np.ndarray`
        Points from the derived distribution to return. For each point, Y, in
        ``input_data``, we determine the value at which a cumulative
        probability of Y is achieved. As a result, all values in
        ``input_data`` must be in the range [0, 1]. Hence if you want a random
        sample from the derived skewed normal, simply make ``input_data``
        equal to a random sample of the uniform distribution [0, 1]

    Returns
    -------
    :obj:`np.ndarray`
        Points sampled from the derived skewed normal distribution based on
        ``input_data``
    """
    ratio = (upper - median) / (median - lower)
    ratio_gte_one = ratio >= 1

    skewed_normal_width = _get_skewed_normal_width(ratio, conf)

    factor_denominator = np.exp(skewed_normal_width * norm.ppf(0.5 * (1 + conf))) - 1

    if ratio_gte_one:
        factor = (upper - median) / factor_denominator
    else:
        factor = (lower - median) / factor_denominator

    shift = median - factor

    input_data = np.array(input_data)
    if ratio_gte_one:
        res_inp = input_data
    else:
        res_inp = 1 - input_data

    res = np.exp(norm.ppf(res_inp) * skewed_normal_width) * factor + shift

    return res


def sample_multivariate_skewed_normal(configuration, size, cor=None):
    r"""
    Sample multi-variate skewed normal distribution

    Following [Meinshausen et al. (2009)](https://doi.org/10.1038/nature08017),
    a skewed normal is defined as follows:
    "A distribution X, is skewed normal, if :math:`\\log(X + C)`, where
    :math:`C` is a constant, is a normal distribution with variance
    :math:`\\sigma^2` and mean :math:`\\mu`". The skewed normal allows us to
    create distributions which match arbitrary median and likely ranges (with
    associated confidence as a percentage), as is often used by the IPCC. A
    multivariate skewed normal is constructed such that each marginal
    distribution is a skewed normal and the overall distribution
    can have a non-identity correlation matrix.

    Parameters
    ----------
    configuration : :obj:`pd.DataFrame`
        Configuration for the sampling. Each column must represent a dimension to
        be sampled. The rows must be
        ``["median", "upper", "lower", "conf"]``. The rows represent the
        characteristics of each marginal distribution. The median is the median
        of each marginal distribution. ``"conf"`` represents the confidence
        associated with each of the intervals defined by ``"lower"`` and
        ``"upper"`` (e.g. 0.66 would mean that [lower, upper] defines the 66%
        confidence range).

    size : int
        Number of points to sample from the multivariate skewed normal

    cor : array[float]
        Correlation matrix between different dimensions in `configuration`.
        `cor` must be a square, symmetric matrix with size n x n, where n is
        the number of columns in ``configuration``. The element in row i and
        column j represents the correlation between the dimension in column i of
        ``configuration`` and the dimension in column j of ``configuration``. The
        correlations must be normalised i.e. the maximum value of each element
        is 1 and the minimum value is -1. As a result of the normalisation, the
        diagonals of ``cor`` must all be equal to 1.

    Returns
    -------
    :obj:`pd.DataFrame`
        Points sampled from the derived multivariate skewed normal
        distribution. Each row is a sampled point (so there will be ``size``
        row in the output). The columns match the columns in ``configuration``.

    Raises
    ------
    ValueError
        ``configuration`` contains any nans, distribution ranges are
        incorrectly specified (e.g. medians > upper ends of the
        intervals) or the correlation matrix is incorrectly normalised.

    KeyError
        ``configuration`` is missing one of the rows:
        ``["median", "upper", "lower", "conf"]``
    """
    if configuration.isnull().any().any():
        raise ValueError(
            f"`configuration` must not contain nan, received: {configuration}"
        )

    res_array = _sample_multivariate_skewed_normal_from_arrays(
        configuration.loc["median", :],
        configuration.loc["lower", :],
        configuration.loc["upper", :],
        configuration.loc["conf", :],
        size,
        cor,
    )

    res = pd.DataFrame(res_array, columns=configuration.columns)

    return res


def _check_input_ordering(a, b, a_name, b_name):
    a_gt_b = a > b
    if not np.all(a_gt_b):
        raise ValueError(
            f"{a_name} must be greater than {b_name}. "
            f"Received {a_name}: {a} and {b_name}: {b}. "
            f"Entries where {a_name} are greater than {b_name}: {a_gt_b}"
        )


def _check_correlation_matrix(cor):
    if not np.allclose(np.diag(cor), 1):
        raise ValueError(
            "The correlation matrix must be normalised i.e. all "
            "diagonal elements should be equal to 1. "
            f"Received: {cor}"
        )

    if not np.all(np.abs(cor) <= 1):
        raise ValueError(
            "The correlation matrix must be normalised i.e. all "
            "off-diagonal elements should have magnitude <=1. "
            f"Received: {cor}"
        )


def _sample_multivariate_skewed_normal_from_arrays(
    medians, lowers, uppers, confs, size, cor
):
    medians = np.atleast_1d(np.asarray(medians))
    lowers = np.atleast_1d(np.asarray(lowers))
    uppers = np.atleast_1d(np.asarray(uppers))
    confs = np.atleast_1d(np.asarray(confs))

    _check_input_ordering(uppers, medians, "uppers", "medians")
    _check_input_ordering(medians, lowers, "medians", "lowers")

    if cor is None:
        cor = np.eye(medians.shape[0])
    else:
        cor = np.asarray(cor)
        _check_correlation_matrix(cor)

    ratios = (uppers - medians) / (medians - lowers)
    ratios_gte_one = ratios >= 1

    skewed_normal_widths = np.asarray(
        [_get_skewed_normal_width(r, c) for r, c in zip(ratios, confs)]
    )

    factors_denominator = np.exp(skewed_normal_widths * norm.ppf(0.5 * (1 + confs))) - 1

    factors = (lowers - medians) / factors_denominator
    factors[ratios_gte_one] = ((uppers - medians) / factors_denominator)[ratios_gte_one]

    shifts = medians - factors

    inp = multivariate_normal(np.zeros_like(medians), cor).rvs(size)
    if len(inp.shape) == 1:
        inp = inp[:, np.newaxis]

    inp[:, np.where(~ratios_gte_one)] *= -1

    # Each marginal distribution has a distribution X, such that
    # log(X + C) is a normal distribution with mean mu and variance sigma^2
    # This can be seen by re-arranging the below to
    # inp * skewed_normal_widths + factors = log(res - shifts)
    # and noting that inp is a normal distribution along each marginal axis
    res = np.exp(inp * skewed_normal_widths) * factors + shifts

    return res



"""
Metric calculations used in RCMIP
"""
from .airborne_fraction import (  # noqa
    CalculatorAirborneFraction18501920,
    CalculatorAirborneFraction18501990,
)
from .tcr import CalculatorTCR  # noqa
from .tcre import CalculatorTCRE  # noqa



"""
Custom errors defined within pyrcmip
"""


class ProtocolConsistencyError(ValueError):
    """
    Inconsistency between input data and the RCMIP protocol
    """


class NoDataForMetricError(ValueError):
    """
    No data available to calculate the given metric
    """



"""
Base class for metric calculations
"""
from abc import ABC


class Calculator(ABC):
    """
    Base class for metric calculations
    """

    @classmethod
    def can_calculate_metric(cls, metric):
        """
        Decide whether the input metric can be calculated or not

        Parameters
        ----------
        metric : str
            Metric to check

        Returns
        -------
        bool
            If ``True``, the metric can be calculated. Otherwise, it cannot.
        """
        return cls._can_calculate_metric(metric)

    @classmethod
    def _can_calculate_metric(cls, metric):  # noqa: D401
        """
        Internal handling of whether a metric can be calculated or not, to be implemented in sub-classes of :obj:`Calculator`
        """
        raise NotImplementedError

    @classmethod
    def calculate_metric(
        cls, assessed_ranges, res_calc, norm_period, evaluation_period, unit
    ):
        """
        Calculate metric

        Parameters
        ----------
        assessed_ranges : :obj:`pyrcmip.assessed_ranges.AssessedRanges`
            Assessed ranges instance

        res_calc : :obj:`scmdata.ScmRun`
            Results from which the metric is to be derived

        norm_period : list
            Years to use for normalising the data before calculating the metric

        evaluation_period : list
            Years to use when evaluating the metric

        unit : str
            Unit in which the metric should be returned

        Returns
        -------
        :obj:`pd.DataFrame`
            Metric values with other relevant model metadata

        Raises
        ------
        NoDataForMetricError
            No data is available to calculate the given metric

        DimensionalityError
            The units of the data cannot be converted to the desired units or
            the units of the data are incompatible with the metric calculation
        """
        return cls._calculate(
            assessed_ranges, res_calc, norm_period, evaluation_period, unit
        )

    @classmethod
    def _calculate(
        cls, assessed_ranges, res_calc, norm_period, evaluation_period, unit
    ):  # noqa: D401
        """
        Internal handling of calculations, to be implemented in sub-classes of :obj:`Calculator`
        """
        raise NotImplementedError

    @staticmethod
    def _check_norm_period_evaluation_period(
        norm_period, expected_norm_period, evaluation_period, expected_evaluation_period
    ):
        if evaluation_period is not None and (
            len(evaluation_period) != 1
            or evaluation_period[0] != expected_evaluation_period
        ):
            raise NotImplementedError(
                "Evaluation period other than [{}], input: {}".format(
                    expected_evaluation_period, evaluation_period
                )
            )

        if norm_period is not None and (
            len(norm_period) != 1 or norm_period[0] != expected_norm_period
        ):
            raise NotImplementedError(
                "Normalisation period other than [{}], input: {}".format(
                    expected_norm_period, norm_period
                )
            )


class _CalculatorTCRTCREBase(Calculator):
    @classmethod
    def _get_normed_res_calc(
        cls, assessed_ranges, res_calc, norm_period, evaluation_period
    ):
        expected_norm_period = 1850

        cls._check_norm_period_evaluation_period(
            norm_period=norm_period,
            expected_norm_period=expected_norm_period,
            evaluation_period=evaluation_period,
            expected_evaluation_period=1920,
        )

        res_calc_normed = assessed_ranges._get_normed_res_calc(
            res_calc, norm_period=expected_norm_period
        )

        return res_calc_normed



"""
Calculation of the transient climate response (TCR)
"""
from .base import _CalculatorTCRTCREBase
from .utils import _time_mean


class CalculatorTCR(_CalculatorTCRTCREBase):
    """
    Calculator of the transient climate response (TCR)
    """

    @classmethod
    def _can_calculate_metric(cls, metric):
        return metric == "Transient Climate Response"

    @classmethod
    def _calculate(
        cls, assessed_ranges, res_calc, norm_period, evaluation_period, unit
    ):
        res_calc_normed = cls._get_normed_res_calc(
            assessed_ranges, res_calc, norm_period, evaluation_period
        )

        out = _time_mean(
            res_calc_normed.filter(
                scenario="1pctCO2",
                variable="Surface Air Temperature Change",
                unit=unit,
                region="World",
                year=1920,
            )
        )

        return out



"""
Handling of assessed ranges
"""
import warnings
from collections import defaultdict

import matplotlib.cbook as cbook
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import tqdm.autonotebook as tqdman
from matplotlib.gridspec import GridSpec
from scmdata import run_append

from .metric_calculations.utils import _regression_slope, _time_mean
from .plotting import CLIMATE_MODEL_PALETTE
from .stats import get_skewed_normal


class AssessedRanges:
    """
    Class for handling assessed ranges and performing operations with them.

    For example, getting values for specific metrics and plotting results against
    assessed ranges.
    """

    metric_column = "RCMIP name"
    """str: Name of the column which holds the names of the metrics being assessed"""

    assessed_range_label = "assessed range"
    """str: String used for labelling assessed ranges (in plots, dataframes etc.)"""

    def __init__(self, db):
        """
        Initialise

        Parameters
        ----------
        db : :obj:`pd.DataFrame`
            Assessed ranges data
        """
        metric_names = db[self.metric_column]
        if metric_names.duplicated().any():
            duplicates = metric_names[metric_names.duplicated()].tolist()
            raise ValueError(
                "Your assessed ranges do not have unique values for `RCMIP "
                "name`. Duplicates: {}".format(duplicates)
            )

        self.db = db.copy()

    def head(self, n=5):
        """
        Get head of ``self.db``

        Parameters
        ----------
        n : int
            Number of rows to return

        Returns
        -------
        :obj:`pd.DataFrame`
            Head of ``self.db``
        """
        return self.db.head(n)

    def tail(self, n=5):
        """
        Get tail of ``self.db``

        Parameters
        ----------
        n : int
            Number of rows to return

        Returns
        -------
        :obj:`pd.DataFrame`
            Tail of ``self.db``
        """
        return self.db.tail(n)

    def get_col_for_metric(self, metric, col):
        """
        Get value of column for a given metric (i.e. RCMIP name)

        Parameters
        ----------
        metric : str
            Metric whose values we want to look up

        col : str
            Column whose values we want (e.g. "RCMIP scenario")

        Returns
        -------
        str
            The value in the column

        Raises
        ------
        ValueError
            The metric could not be found in ``self.db``

        KeyError
            The column could not be found in ``self.db``
        """
        metric_row = self.db[self.metric_column] == metric
        if not metric_row.any():
            raise ValueError("Metric `{}` not found".format(metric))

        try:
            out = self.db.loc[metric_row, col].tolist()
        except KeyError:
            raise KeyError("Column `{}` not found".format(col))

        return out[0]

    def get_col_for_metric_list(self, metric, col, delimeter=","):
        """
        Get value of column for a given metric (i.e. RCMIP name), split using a delimeter

        Parameters
        ----------
        metric : str
            Metric whose values we want to look up

        col : str
            Column whose values we want (e.g. "RCMIP scenario")

        delimeter : str
            Delimeter used to split ``col``'s values

        Returns
        -------
        list
            List of values, derived by splitting

        Raises
        ------
        TypeError
            The found values are not a string (i.e. cannot be split by a
            delimiter)
        """
        out = self.get_col_for_metric(metric, col)

        try:
            return [v.strip() for v in out.split(delimeter)]
        except (TypeError, AttributeError):
            raise TypeError("Cannot split `{}` of type `{}`".format(out, type(out)))

    def get_variables_regions_scenarios_for_metric(self, metric, single_value=True):
        """
        Get variables, regions and scenarios required to calculate a given metric

        Parameters
        ----------
        metric : str
            Metric for which to get values

        Returns
        -------
        dict
            Dictionary containing required variables, regions and scenarios
        """
        return {
            id_str: self.get_col_for_metric_list(metric, "RCMIP {}".format(id_str))
            for id_str in ("variable", "region", "scenario")
        }

    def get_norm_period_evaluation_period(self, metric):
        """
        Get normalisation and evaluation period for a given metric

        Parameters
        ----------
        metric : str
            Metric for which to get normalisation and evaluation periods

        Returns
        -------
        norm_period, evaluation_period
            Normalisation period and evaluation period. Each return value is a
            range of years which define the relevant period. If there is no
            period supplied, ``None`` is returned. For example, if the
            evaluation period is 1961-1990 and there is no reference period,
            then ``None, range(1961, 1990 + 1)`` is returned.

        Raises
        ------
        ValueError
            A period could not be resolved because it is ambiguous i.e. it has
            nan for the start/end of the period while the other value is not
            nan.
        """
        norm_period = self._check_period_values(
            self.get_col_for_metric(metric, "norm_period_start"),
            self.get_col_for_metric(metric, "norm_period_end"),
            "norm_period",
        )

        evaluation_period = self._check_period_values(
            self.get_col_for_metric(metric, "evaluation_period_start"),
            self.get_col_for_metric(metric, "evaluation_period_end"),
            "evaluation_period",
        )

        return norm_period, evaluation_period

    @staticmethod
    def _check_period_values(start, end, period_name):
        start_is_nan = np.isnan(start)
        end_is_nan = np.isnan(end)

        if start_is_nan and end_is_nan:
            return None

        if not (start_is_nan or end_is_nan):
            return range(int(start), int(end) + 1)

        if start_is_nan:
            raise ValueError(
                "Ambiguous {}, start is nan and end is {}".format(period_name, end)
            )

        raise ValueError(
            "Ambiguous {}, start is {} and end is nan".format(period_name, start)
        )

    def check_norm_period_evaluation_period_against_data(
        self, norm_period, evaluation_period, data
    ):
        """
        Check the normalisation and evaluation periods against the data

        Parameters
        ----------
        norm_period : None or range(int, int)
            Normalisation period to check. If ``None``, no check is performed.

        evaluation_period : None or range(int, int)
            Evaluation period to check. If ``None``, no check is performed.

        data : :obj:`scmdata.ScmRun`
            Data to check

        Raises
        ------
        ValueError
            The data is incompatible with the periods (e.g. the normalisation
            period begins before the data begins).
        """
        res_min_year = data["year"].min()
        res_max_year = data["year"].max()

        self._check_period_against_data(
            norm_period, "normalisation period", res_min_year, res_max_year
        )
        self._check_period_against_data(
            evaluation_period, "evaluation period", res_min_year, res_max_year
        )

    @staticmethod
    def _check_period_against_data(period, period_id, res_min_year, res_max_year):
        if period is not None:
            if res_min_year > min(period):
                raise ValueError(
                    "Res min. year: {} is greater than {} min.: {}".format(
                        res_min_year, period_id, min(period)
                    )
                )

            if res_max_year < max(period):
                raise ValueError(
                    "Res max. year: {} is less than {} max.: {}".format(
                        res_max_year, period_id, max(period)
                    )
                )

    def get_assessed_range_for_boxplot(self, metric, n_to_draw=2 * 10**4):
        """
        Get assessed range for a box plot

        This converts the assessed range from IPCC language (very likely,
        likely, central) into a distribution of values, based on
        :func:`pyrcmip.stats.get_skewed_normal`.

        Parameters
        ----------
        metric : str
            Metric for which to get assessed range distribution

        n_to_draw : int
            Number of points to include in the returned distribution

        Returns
        -------
        :obj:`pd.DataFrame`
            :obj:`pd.DataFrame` with ``n_to_draw`` rows, each of which
            contains a drawn value for ``metric``. The returned values are put
            in a column whose name is equal to the value of ``metric``. We
            also return a ``"unit"`` column and a ``"Source"`` column. The
            ``"Source"`` column is filled with ``self.assessed_range_label``.
            Note that if the central value is nan, the entire distribution
            will simply be filled with nan.
        """
        values = self._get_assessed_range_values_for_metric(metric)

        if np.isnan(values["very_likely__lower"]):
            lower = values["likely__lower"]
            upper = values["likely__upper"]
            conf = 0.66
        else:
            lower = values["very_likely__lower"]
            upper = values["very_likely__upper"]
            conf = 0.9

        if np.isnan(values["central"]):
            out = np.zeros(n_to_draw) * np.nan
        else:
            # generate array which will plot as desired
            out = get_skewed_normal(
                median=values["central"],
                lower=lower,
                upper=upper,
                conf=conf,
                input_data=np.random.random(n_to_draw),
            )

        ar_for_boxplot = pd.DataFrame(out, columns=[metric])
        ar_for_boxplot["Source"] = self.assessed_range_label
        ar_for_boxplot["unit"] = self.get_col_for_metric(metric, "unit")

        return ar_for_boxplot

    def _get_assessed_range_values_for_metric(self, metric):
        value_cols = [
            "very_likely__lower",
            "likely__lower",
            "central",
            "likely__upper",
            "very_likely__upper",
        ]
        values = {k: float(self.get_col_for_metric(metric, k)) for k in value_cols}

        return values

    def get_results_summary_table_for_metric(self, metric, model_results):
        """
        Get results summary table for a given metric

        Parameters
        ----------
        metric : str
            Metric for which to get the summary table

        model_results : :obj:`pd.DataFrame`
            :obj:`pd.DataFrame` containing the model results. It must have at
            least the following columns: ``"climate_model", "value"``.

        Returns
        -------
        :obj:`pd.DataFrame`
            :obj:`pd.DataFrame` containing a summary of the results. The
            percentage difference is calculated as
            ``(model_value - assessed_value) / np.abs(assessed_value) * 100``.
        """
        model_results_metric_row = model_results[self.metric_column] == metric
        if not model_results_metric_row.any():
            raise KeyError("No model results for metric: {}".format(metric))

        model_results_metric = model_results.loc[model_results_metric_row, :]

        assessed_ranges_unit = self.get_col_for_metric(metric, "unit")
        model_results_metric_unit = model_results_metric["unit"].unique().tolist()
        if (
            len(model_results_metric_unit) > 1
            or model_results_metric_unit[0] != assessed_ranges_unit
        ):
            error_msg = "Units mismatch. Assessed range units: {}, model results unit(s): {}".format(
                assessed_ranges_unit, model_results_metric_unit
            )
            raise AssertionError(error_msg)

        summary_table = defaultdict(list)
        assessed_range_quantiles = {
            "very_likely__lower": 0.05,
            "likely__lower": 0.17,
            "central": 0.5,
            "likely__upper": 0.83,
            "very_likely__upper": 0.95,
        }
        for label, quantile in tqdman.tqdm(
            assessed_range_quantiles.items(), leave=False
        ):
            assessed_value = self.get_col_for_metric(metric, label)
            if np.isnan(assessed_value):
                continue

            for climate_model, df in model_results_metric.groupby("climate_model"):
                model_quantile = df["value"].quantile(quantile)

                summary_table["climate_model"].append(climate_model)
                summary_table["assessed_range_label"].append(label)
                summary_table["assessed_range_value"].append(assessed_value)
                summary_table["climate_model_value"].append(model_quantile)
                summary_table["percentage_difference"].append(
                    (model_quantile - assessed_value) / np.abs(assessed_value) * 100
                )

        summary_table = pd.DataFrame(summary_table)
        summary_table[self.metric_column] = metric
        summary_table["unit"] = assessed_ranges_unit

        return summary_table

    def plot_metric_and_results(self, metric, model_results, axes=None, palette=None):
        """
        Plot our parameterisation of the metric's distribution and the model results

        This produces a two-panel plot, the top panel has the distributions,
        the bottom panel has box and whisker plots (with the boxes and
        whiskers adjusted to match the IPCC calibrated likelihood language).

        Parameters
        ----------
        metric : str
            Metric to plot

        model_results : :obj:`pd.DataFrame`
            :obj:`pd.DataFrame` with the model results. Should be of the form
            returned by :meth:`calculate_metric_from_results`.

        axes : (:obj:`matplotlib.axes.SubplotBase`, :obj:`matplotlib.axes.SubplotBase`)
            Axes on which to make the plots. Must be two-panels.

        palette : dict[str, str]
            Colours to use for the different climate models and assessed ranges

        Returns
        -------
        (:obj:`matplotlib.axes.SubplotBase`, :obj:`matplotlib.axes.SubplotBase`)
            Axes on which the plot was made

        Raises
        ------
        AssertionError
            ``axes`` doesn't have a length equal to two
        """
        if axes is None:
            axes = self._get_evaluation_axes()
        else:
            assert len(axes) == 2, "Must pass two axes, received {}".format(len(axes))

        return self._plot(
            metric, model_results, axs=axes, box_only=False, palette=palette
        )

    @staticmethod
    def _get_evaluation_axes():
        fig = plt.figure(figsize=(12, 7))
        gs = GridSpec(2, 1, width_ratios=[1], height_ratios=[3, 2])
        ax0 = fig.add_subplot(gs[0])
        ax1 = fig.add_subplot(gs[1], sharex=ax0)
        axes = [ax0, ax1]

        return axes

    def plot_metric_and_results_box_only(
        self, metric, model_results, ax=None, palette=None
    ):
        """
        Plot box and whisker plots of the metric's distribution and the model results

        The box and whisker plots have the boxes and whiskers adjusted to
        match the IPCC calibrated likelihood language).

        Parameters
        ----------
        metric : str
            Metric to plot

        model_results : :obj:`pd.DataFrame`
            :obj:`pd.DataFrame` with the model results. Should be of the form
            returned by :meth:`calculate_metric_from_results`.

        axes : :obj:`matplotlib.axes.SubplotBase`
            Axis on which to make the plot

        palette : dict[str, str]
            Colours to use for the different climate models and assessed ranges

        Returns
        -------
        :obj:`matplotlib.axes.SubplotBase`
            Axes on which the plot was made
        """
        if ax is None:
            ax = plt.figure().add_subplot(111)

        return self._plot(metric, model_results, axs=ax, box_only=True, palette=palette)

    def _plot(self, metric, model_results, axs, box_only, palette=None):
        if palette is None:
            palette = CLIMATE_MODEL_PALETTE

        box_plot_df = self._get_box_plot_df(metric, model_results)
        unit = box_plot_df["unit"].unique().tolist()
        assert len(unit) == 1, unit
        unit = unit[0]

        if box_only:
            box_ax = axs
        else:
            box_ax = axs[1]

        box_plot_stats = self._get_box_whisker_stats_custom_quantiles(
            box_plot_df, metric, box_quantiles=(17, 83), whisker_quantiles=(5, 95)
        )
        box_ax.bxp(
            box_plot_stats,
            showfliers=False,
            vert=box_only,
        )

        xlim = box_ax.get_xlim()

        if box_only:
            box_ax.set_ylabel(unit)
            box_ax.set_title(metric)
            box_ax.grid(axis="y")
        else:
            box_ax.set_xlabel(unit)
            box_ax.set_title("")
            box_ax.grid(axis="x")

            pdf_ax = axs[0]
            sns.histplot(
                data=box_plot_df,
                x=metric,
                bins=50,
                stat="probability",
                common_norm=False,
                kde=True,
                hue="Source",
                palette=palette,
                ax=pdf_ax,
                legend=True,
            )

            pdf_ax.set_title(metric)
            pdf_ax.set_xlim(xlim)
            pdf_ax.set_xlabel("")

        plt.suptitle("")

        return axs

    def _get_box_plot_df(self, metric, model_results):
        model_results_metric = model_results.loc[
            model_results[self.metric_column] == metric, :
        ]
        res = model_results_metric.pivot_table(
            values="value",
            columns=self.metric_column,
            index=list(set(model_results.columns) - {"value", self.metric_column}),
        ).reset_index()

        res["Source"] = res["climate_model"]

        assessed_ranges = self.get_assessed_range_for_boxplot(metric)

        box_plot_df = pd.concat([assessed_ranges, res])[[metric, "Source", "unit"]]

        unit = box_plot_df["unit"].unique().tolist()
        if len(unit) != 1:
            raise AssertionError("Not a single unit, found: {}".format(unit))

        return box_plot_df

    def _get_box_whisker_stats_custom_quantiles(
        self, model_df, metric, box_quantiles=(17, 83), whisker_quantiles=(5, 95)
    ):
        """
        Note that using box_quantiles or whisker_quantiles other than the default
        would imply changing the IPCC likelihood language. As a result, we do not
        currently expose these keyword arguments via any "public" APIs.
        """
        self._check_box_quantiles(box_quantiles, "box")
        self._check_box_quantiles(whisker_quantiles, "whisker")

        assessment = self._get_assessed_range_values_for_metric(metric)

        stats = {}
        for source, df in model_df.groupby("Source"):
            data = df[metric].values.squeeze()

            stats[source] = cbook.boxplot_stats(data, labels=[source])[0]

            if source == self.assessed_range_label:
                _assessment = assessment
            else:
                # Make a dictionary with all nan so all calculations are triggered.
                # This is slightly more expensive but ensures consistency.
                _assessment = {k: np.nan for k in assessment.keys()}

            stats[source] = self._fill_for_matplotlib_based_on_assessment(
                stats[source],
                assessed_vals=_assessment,
                assessed_keys=("central",),
                assessed_implied_quantiles=(50,),
                matplotlib_keys=("med",),
                values=data,
            )

            stats[source] = self._fill_for_matplotlib_based_on_assessment(
                stats[source],
                assessed_vals=_assessment,
                assessed_keys=("likely__lower", "likely__upper"),
                assessed_implied_quantiles=box_quantiles,
                matplotlib_keys=("q1", "q3"),
                values=data,
            )

            stats[source] = self._fill_for_matplotlib_based_on_assessment(
                stats[source],
                assessed_vals=_assessment,
                assessed_keys=("very_likely__lower", "very_likely__upper"),
                assessed_implied_quantiles=whisker_quantiles,
                matplotlib_keys=("whislo", "whishi"),
                values=data,
            )

        return list(stats.values())

    @staticmethod
    def _check_box_quantiles(quantiles, label_str):
        if quantiles[0] >= quantiles[1]:
            raise AssertionError(
                "{} quantiles must be ordered from lower to higher, input: {}".format(
                    label_str, quantiles
                )
            )

    @staticmethod
    def _fill_for_matplotlib_based_on_assessment(
        holder,
        assessed_vals,
        assessed_keys,
        assessed_implied_quantiles,
        matplotlib_keys,
        values,
    ):
        assessed_vals_of_interest = {
            k: v for k, v in assessed_vals.items() if k in assessed_keys
        }

        nan_assessed_values = np.isnan(list(assessed_vals_of_interest.values()))

        if not nan_assessed_values.any():
            for matplotlib_key, assessed_key in zip(matplotlib_keys, assessed_keys):
                holder[matplotlib_key] = assessed_vals[assessed_key]

        elif nan_assessed_values.all():
            for matplotlib_key, quantile in zip(
                matplotlib_keys, assessed_implied_quantiles
            ):
                holder[matplotlib_key] = np.percentile(values, quantile)

        else:  # pragma: no cover
            raise NotImplementedError(
                "One half of the assessed range is nan and the other is not? ({})".format(
                    assessed_vals
                )
            )

        return holder

    def calculate_metric_from_results(self, metric, res_calc, custom_calculators=None):
        """
        Calculate metric values from results

        Parameters
        ----------
        metric : str
            Metric for which to calculate results

        res_calc : :obj:`scmdata.ScmRun`
            Results to use for the calculation

        custom_calculators : tuple(:obj:`pyrcmip.metric_calculations.base.Calculator`)
            Custom calculators to use for calculating metrics which require a
            custom calculation

        Returns
        -------
        :obj:`pd.DataFrame`
            :obj:`pd.DataFrame` containing the calculated metric values
            alongside other relevant metadata

        Raises
        ------
        ValueError
            Data required to calculate the metric is not available
        """
        filters = self.get_variables_regions_scenarios_for_metric(metric)
        res_calc = res_calc.filter(**filters, log_if_empty=False)
        if res_calc.empty:
            raise ValueError(
                "{} requires {} but it is not in the results".format(metric, filters)
            )

        eval_method = self.get_col_for_metric(metric, "RCMIP evaluation method")

        (norm_period, evaluation_period) = self.get_norm_period_evaluation_period(
            metric
        )
        self.check_norm_period_evaluation_period_against_data(
            norm_period, evaluation_period, res_calc
        )

        unit = self.get_col_for_metric(metric, "unit")

        if eval_method == "mean":
            res_calc_normed = self._get_normed_res_calc(res_calc, norm_period)
            res_calc_normed = res_calc_normed.convert_unit(unit)

            derived = _time_mean(res_calc_normed.filter(year=evaluation_period))

        elif eval_method == "regression-slope":
            res_calc_normed = self._get_normed_res_calc(res_calc, norm_period)

            derived = _regression_slope(
                res_calc_normed.filter(year=evaluation_period), unit
            )

        elif eval_method == "custom":
            if custom_calculators is None:
                custom_calculators = []

            calculator = [
                c for c in custom_calculators if c.can_calculate_metric(metric)
            ]
            if not calculator:
                raise ValueError("No custom calculator for {}".format(metric))

            if len(calculator) > 1:
                raise ValueError(
                    "More than one available calculator for metric {}, matched: {}".format(
                        metric, calculator
                    )
                )

            derived = calculator[0].calculate_metric(
                self, res_calc, norm_period, evaluation_period, unit
            )

        else:
            raise NotImplementedError(eval_method)

        derived = derived.to_frame().reset_index()
        derived[self.metric_column] = metric

        return derived

    @staticmethod
    def _get_normed_res_calc(res_calc, norm_period):
        if norm_period is not None:
            res_calc_normed = res_calc.relative_to_ref_period_mean(year=norm_period)
        else:
            res_calc_normed = res_calc.copy()

        return res_calc_normed

    def plot_against_results(
        self,
        results_database,
        climate_models=["*"],
        custom_calculators=None,
        palette=None,
    ):
        """
        Calculate metric values from results, compare and plot against assessed ranges

        Parameters
        ----------
        metric : str
            Metric for which to calculate results

        results_database : :obj:`pyrcmip.database.DataBase`
            Database from which to load results

        climate_models : list[str]
            Climate models to calculate results for

        custom_calculators : tuple(:obj:`pyrcmip.metric_calculations.base.Calculator`)
            Custom calculators to use for calculating metrics which require a
            custom calculation

        palette : dict[str, str]
            Colours to use for the different climate models and assessed
            ranges when plotting

        Returns
        -------
        :obj:`pd.DataFrame`
            :obj:`pd.DataFrame` containing a dataframe based on concatenating
            the results from calling
            :meth:`get_results_summary_table_for_metric` for each metric.
        """
        summary_table = []
        for metric in tqdman.tqdm(self.db[self.metric_column].tolist()):
            eval_method = self.get_col_for_metric(metric, "RCMIP evaluation method")
            if isinstance(eval_method, float) and np.isnan(eval_method):
                warnings.warn("No evaluation method for {}".format(metric))
                continue

            var_region_scen = self.get_variables_regions_scenarios_for_metric(metric)
            variables = var_region_scen["variable"]
            regions = var_region_scen["region"]
            scenarios = var_region_scen["scenario"]

            try:
                # TODO: upgrade the database so we can specify the climate models to load too
                res_calc = run_append(
                    [
                        results_database.load_data(
                            climate_model=cm, variable=v, region=r, scenario=s
                        )
                        for cm in climate_models
                        for v in variables
                        for r in regions
                        for s in scenarios
                    ]
                ).time_mean("AC")
            except IndexError:
                print("No data for: {}".format(metric))
                continue

            # if we need better control, use a meta argument with this docstring:
            """
            meta : list[str]
                Meta columns to use when performing calculations (all other meta
                columns are dropped)
            """
            # and the code below
            """
            if meta is not None:
                meta_to_drop = set(res_calc.meta.columns) - set(meta)
                res_calc = res_calc.drop_meta(meta_to_drop, inplace=False)
            """

            derived = self.calculate_metric_from_results(
                metric, res_calc, custom_calculators=custom_calculators
            )

            summary_table_metric = self._plot_pdf_and_box_and_get_summary_table(
                metric, derived, palette=palette
            )

            summary_table.append(summary_table_metric)

        summary_table = pd.concat(summary_table).reset_index(drop=True)

        return summary_table

    def plot_model_reported_against_assessed_ranges(self, model_reported, palette=None):
        """
        Compare and plot model reported results against assessed ranges

        Parameters
        ----------
        model_reported : :obj:`pd.DataFrame`
            :obj:`pd.DataFrame` of the same format as the result of
            :meth:`calculate_metric_from_results`

        palette : dict[str, str]
            Colours to use for the different climate models and assessed
            ranges when plotting

        Returns
        -------
        :obj:`pd.DataFrame`
            :obj:`pd.DataFrame` containing a dataframe based on concatenating
            the results from calling
            :meth:`get_results_summary_table_for_metric` for each metric
        """
        summary_table = []

        for label, df in model_reported.groupby(self.metric_column):
            summary_table_metric = self._plot_pdf_and_box_and_get_summary_table(
                label, df, palette=palette
            )

            summary_table.append(summary_table_metric)

        summary_table = pd.concat(summary_table).reset_index(drop=True)

        return summary_table

    def _plot_pdf_and_box_and_get_summary_table(
        self, metric, derived, show=True, palette=None
    ):
        axes = self._get_evaluation_axes()

        axes = self.plot_metric_and_results(metric, derived, axes=axes, palette=palette)
        plt.tight_layout()

        ax = plt.figure(figsize=(6, 4)).add_subplot(111)
        self.plot_metric_and_results_box_only(metric, derived, ax=ax, palette=palette)

        plt.tight_layout()
        if show:
            plt.show()

        summary_table_metric = self.get_results_summary_table_for_metric(
            metric, derived
        )

        return summary_table_metric



"""
Input and output handling
"""

import contextlib
import gzip
import io
import logging
import os
import os.path
import tempfile

import click
import pandas as pd
import xarray as xr
from botocore.exceptions import ClientError
from scmdata.run import ScmRun, run_append

LOGGER = logging.getLogger(__name__)

CSV_COLUMN_MAP = {
    "Climate Model": "climate_model",
    "Climate Model Name": "climate_model_name",
    "Climate Model Version": "climate_model_version",
    "Climate Model Configuration Label": "climate_model_configuration_label",
    "Climate Model Configuration Description": "climate_model_configuration_description",
    "Project": "project",
    "Name of Person": "name_of_person",
    "Literature Reference": "literature_reference",
}


def ensure_dir_exists(fp):
    """
    Ensure directory exists

    Parameters
    ----------
    fp : str
        Filepath of which to ensure the directory exists
    """
    dir_to_check = os.path.dirname(fp)
    if not os.path.isdir(dir_to_check):
        LOGGER.info("Creating {}".format(dir_to_check))
        os.makedirs(dir_to_check)


def read_results_submission(results):
    """
    Read results submission

    Parameters
    ----------
    results : str or list of str
        Files to read in. All files to be read should be formatted as csv or
        xlsx files following the formatting defined in the template
        spreadsheet.

    Returns
    -------
    :obj:`scmdata.ScmRun`
        Results read in from the submission(s)
    """
    if isinstance(results, str):
        results = [results]

    db = []
    for rf in results:
        LOGGER.info("Reading %s", rf)

        if rf.endswith(".nc"):
            LOGGER.info("Assuming netCDF format")
            loaded = ScmRun.from_nc(rf)
        else:
            if rf.endswith(".xlsx") or rf.endswith(".xls"):
                LOGGER.info("Assuming excel format")
                loaded = pd.read_excel(rf, sheet_name="your_data", engine="openpyxl")
            else:
                LOGGER.info("Assuming csv format")
                loaded = pd.read_csv(rf)

            LOGGER.debug("Converting all columns to lowercase")
            loaded.columns = loaded.columns.astype(str).str.lower()
            loaded = ScmRun(loaded)

        db.append(ScmRun(loaded))

    LOGGER.info("Joining results together")
    db = run_append(db)

    return db


def read_submission_model_reported(fp):
    """
    Read the model reported component of a submission

    Parameters
    ----------
    fp : str
        Filepath to read

    Returns
    -------
    :obj:`pd.DataFrame`
    """
    out = pd.read_csv(fp)

    return out


def read_submission_model_metadata(fp):
    """
    Read the model metadata component of a submission

    Parameters
    ----------
    fp : str
        Filepath to read

    Returns
    -------
    :obj:`pd.DataFrame`
    """
    LOGGER.info("Reading %s", fp)

    if fp.endswith(".xlsx"):
        LOGGER.info("Assuming excel format")

        meta_df = pd.read_excel(fp, sheet_name="meta_model")

        if len(meta_df) > 3:
            LOGGER.warning(
                "More than one line in 'meta_model' sheet. Only the first will be used."
            )
        elif len(meta_df) < 3:
            error_str = (
                "No model metadata found. Consider filling the 'meta_model' sheet."
            )
            LOGGER.exception(error_str)
            raise click.ClickException(str(error_str))

        meta_df = meta_df.iloc[2:4, :].drop(meta_df.columns[[0]], axis=1)
        meta_df = meta_df.rename(columns=CSV_COLUMN_MAP)

    elif fp.endswith(".csv"):
        LOGGER.info("Assuming CSV format")
        meta_df = pd.read_csv(fp)

        # Assuming the structure is the same as in the .xlsx file
        if "Combination of Climate" in str(meta_df.iloc[0]):
            meta_df = meta_df.iloc[2:, :]

        if len(meta_df) > 1:
            LOGGER.warning(
                "More than one line of metadata answers found. Only the first line will be used."
            )
        elif len(meta_df) < 1:
            error_str = "Metadata file does not have enough rows. Expecting headers + one line of values."
            LOGGER.exception(error_str)
            raise click.ClickException(str(error_str))

        # Make sure the column names are what we expect
        meta_df = meta_df.rename(CSV_COLUMN_MAP)

    return meta_df


def read_submission_model_comments(fp):
    """
    Read the submission comments.

    Parameters
    ----------
    fp : str
        Filepath to read

    Returns
    -------
    :obj:`pd.DataFrame`
    """
    LOGGER.info("Reading %s", fp)

    if fp.endswith(".xlsx"):
        LOGGER.info("Assuming excel format")

        comments_df = pd.read_excel(fp, sheet_name="comments")

        if len(comments_df) < 2:
            error_str = "No submission comments found in supplied file."
            LOGGER.exception(error_str)
            raise click.ClickException(str(error_str))

        comments_df = comments_df.iloc[1:, 2:]

    elif fp.endswith(".csv"):
        LOGGER.info("Assuming CSV format")
        comments_df = pd.read_csv(fp)

        # Assuming the structure is the same as in the .xlsx file
        if "Comments relating" in str(comments_df.iloc[0]):
            comments_df = comments_df.iloc[1:, :].drop(
                [
                    "Comments relating to your submitted data",
                    "Unnamed: 1",
                    "Unnamed: 10",
                ],
                axis=1,
            )

    if len(comments_df) < 1:
        error_str = "No submission comments found in supplied file."
        LOGGER.exception(error_str)
        raise click.ClickException(str(error_str))

    comments_df.columns = comments_df.columns.astype(str).str.lower()
    comments_df = comments_df.rename({"climate model": "climate_model"}, axis=1)

    comments_df = comments_df.dropna(how="all")

    return comments_df


@contextlib.contextmanager
def temporary_file_to_upload(df, max_size=1024, compress=False, file_format="csv"):
    """
    Create a gzipped temporary serialized version of a file to upload

    Attempts to keep the file in memory until it exceeds `max_size`. The file is then stored on-disk
    and cleaned up at the end of the context.

    The temporary location can be overriden using the `TMPDIR` environment variable as per
    https://docs.python.org/3/library/tempfile.html#tempfile.gettempdir

    Notice that this only supports taking an :obj:`scmdata.ScmRun`|:obj:`xarray.Dataset` if the target
    format is CSV (file_format="csv"), and :obj:`xarray.Dataset` if the target format is netCDF
    (file_format="nc").

    Parameters
    ----------
    df : :obj:`scmdata.ScmRun` | :obj:`pd.DataFrame` | :obj:`xarray.Dataset`
        Run to store

    max_size: int or float
        Max size in MB before file is temporarily streamed to disk. Defaults to 1GB

    compress: bool
        whether to compress the file to upload

    file_format: str
        File format for the target file to upload. Allowed values are: "csv" and "nc".

    Returns
    -------
    :obj:`tempfile.SpooledTemporaryFile`
        Open file object ready to be streamed
    """
    # For the time being, this function only supports:
    # SCMrun instance and pandas dataframe if target format is nc
    # Xarray Dataset if target format is nc
    if file_format == "csv" and isinstance(df, (ScmRun, pd.DataFrame)):
        if isinstance(df, ScmRun):
            df = df.timeseries().reset_index()

        if compress:
            buffer = tempfile.SpooledTemporaryFile(max_size=max_size * 1024 * 1024)
            with gzip.GzipFile(mode="w", fileobj=buffer) as gz_file:
                df.to_csv(io.TextIOWrapper(gz_file, "utf8"), index=False)
        else:
            buffer = io.BytesIO(df.to_csv(index=False).encode("utf8"))
    elif file_format == "nc" and isinstance(df, (xr.Dataset)):
        if compress:
            buffer = tempfile.SpooledTemporaryFile(max_size=max_size * 1024 * 1024)
            # write to a real temp file first as a workaround to make this work
            with tempfile.NamedTemporaryFile(suffix=".nc") as tmp:
                df.to_netcdf(tmp.name)
                tmp.seek(0)
                with gzip.GzipFile(mode="w", fileobj=buffer) as gz_file:
                    gz_file.write(tmp.read())
        else:
            buffer = io.BytesIO(df.to_netcdf())
    else:
        raise ValueError(
            f"Unsupported file_format: {file_format} and input format combination"
        )

    buffer.seek(0)

    try:
        yield buffer
    finally:
        buffer.close()


def _upload_file(client, bucket, key, run, compress=True, file_format="csv"):
    """
    Upload a file (CSV or NetCDF) to S3.

    Small files are kept in memory while bigger files are written to disk and automatically cleaned up
    """
    LOGGER.info("Preparing {} for upload".format(key))
    with temporary_file_to_upload(
        run, max_size=1024, compress=compress, file_format=file_format
    ) as fh:
        try:
            if file_format in ("csv", "nc"):
                key += f".{file_format}"
            else:
                raise ValueError(f"Unsupported file_format: {format}")
            if compress:
                key += ".gz"
            LOGGER.info("Uploading {}".format(key))
            client.upload_fileobj(fh, Bucket=bucket, Key=key)
        except ClientError:  # pragma: no cover
            LOGGER.exception("Failed to upload file")
            raise click.ClickException("Failed to upload file")



"""
Utility functions for the CLI tool
"""

import datetime
import logging

import numpy as np
import xarray as xr
from scmdata.run import ScmRun

LOGGER = logging.getLogger(__name__)


def tabular_to_netcdf(data_df, meta_df, version="0.0.0"):
    """
    Convert climate model data (CSV + metadata CSV OR Excel) into NetCDF format.

    Structure:
    - Dimensions: ensemble_member, variable, region, time
    - Attributes: model, unit for each variable + metadata

    Parameters
    ----------
    data_df : :obj:`scmdata.ScmRun` or :obj:`pd.DataFrame`
        Submission data.
    meta_df : :obj:`pd.DataFrame`
        Metadata associated with the submission data.
    version : str
        Version string to include in filename.

    Returns
    -------
    dict[str, xarray.Dataset]
        Dictionary mapping each scenario name to an :class:`xarray.Dataset`
        containing the climate model data and metadata in NetCDF format.
    """
    LOGGER.info("Coverting tabular data to NetCDF file")

    if isinstance(data_df, ScmRun):
        data_df = data_df.timeseries().reset_index()

    # Identify time columns (numeric strings)
    time_cols = [
        col
        for col in data_df.columns
        if isinstance(col, (int, datetime.datetime))
        or (isinstance(col, str) and col.isdigit())
    ]

    time = np.array(
        [
            col.year if isinstance(col, datetime.datetime) else int(col)
            for col in time_cols
        ]
    )

    # Identify climate model
    model_name = data_df["climate_model"].unique()
    if len(model_name) > 1:
        LOGGER.warning(
            "More than one climate model found. Only the first will be used."
        )
    model_name = model_name[0]

    # Extra model metadata
    headers = meta_df.columns
    values = meta_df.iloc[0]
    meta_attrs = dict(zip(headers, values))

    scenario_dict = {}
    # Group by scenario
    for scenario, group in data_df.groupby("scenario"):
        ensemble_members = group["ensemble_member"].unique()
        variables = group["variable"].unique()
        regions = group["region"].unique()

        # Identify IAM for the scenario
        iam_name = group["model"].unique()
        if len(iam_name) > 1:
            LOGGER.warning(
                "More than one IAM for {}. Only the first will be used.".format(
                    scenario
                )
            )
        iam_name = iam_name[0]

        # Prepare empty data array
        data = np.full(
            (len(ensemble_members), len(variables), len(regions), len(time)),
            np.nan,
            dtype=float,
        )

        # Metadata dict for variables
        metadata = {}

        # Fill in data
        for _, row in group.iterrows():
            var = row["variable"]
            em = row["ensemble_member"]
            re = row["region"]

            var_idx = np.where(variables == var)[0][0]
            em_idx = np.where(ensemble_members == em)[0][0]
            re_idx = np.where(regions == re)[0][0]

            data[em_idx, var_idx, re_idx, :] = row[time_cols].to_numpy(dtype=float)
            metadata[var] = {"unit": row["unit"]}

        LOGGER.info(f"Building netcdf file for {scenario}")
        # Build dataset
        ds = xr.Dataset(
            {
                var: (("ensemble_member", "region", "time"), data[:, vi, :, :])
                for vi, var in enumerate(variables)
            },
            coords={
                "variable": variables,
                "ensemble_member": ensemble_members,
                "region": regions,
                "time": time,
            },
        )

        # Add attributes
        for var in variables:
            ds[var].attrs.update(metadata[var])

        ds.attrs["version"] = version
        ds.attrs["scenario"] = scenario
        ds.attrs["model"] = iam_name
        ds.attrs.update(meta_attrs)

        scenario_dict[scenario] = ds

    return scenario_dict



"""
Database of results handling
"""
import glob
import os
import os.path

import pandas as pd
import tqdm.autonotebook as tqdman
from scmdata import ScmRun, run_append

from .io import ensure_dir_exists


class Database:
    """
    On-disk database handler for outputs from SCMs
    """

    def __init__(self, root_dir):
        """
        Initialise the database handler

        Parameters
        ----------
        root_dir : str
            The root directory of the database
        """
        self._root_dir = root_dir

    def __repr__(self):
        return "<utils.scmdata.Database (root_dir: {})>".format(self._root_dir)

    @staticmethod
    def _get_disk_filename(inp):
        return inp.replace("|", "-").replace(" ", "-")

    def save_to_database(self, scmrun):
        """
        Save a set of results to the database

        The results are saved with one file for each
        ``["climate_model", "variable", "region", "scenario", "ensemble_member"]``
        combination.

        Parameters
        ----------
        scmrun : :obj:`scmdata.ScmRun`
            Results to save
        """
        for r in tqdman.tqdm(
            scmrun.groupby(
                ["climate_model", "variable", "region", "scenario", "ensemble_member"]
            ),
            leave=False,
            desc="Saving to database",
        ):
            self._save_to_database_single_file(r)

    def get_out_filepath(
        self, climate_model, variable, region, scenario, ensemble_member=None
    ):
        """
        Get filepath in which data has been saved

        The filepath is the root directory joined with the other information provided. The filepath
        is also cleaned to remove spaces and special characters.

        Parameters
        ----------
        climate_model : str
            Climate model to retrieve data for

        variable : str
            Variable to retrieve data for

        region : str
            Region to retrieve data for

        scenario : str
            Scenario to retrieve data for

        ensemble_member : str or None
            Ensemble member to retrieve data for

        Returns
        -------
        str
            Path in which to save the data. If ``ensemble_member`` is ``None`` then it is not
            included in the filename.
        """
        out_dir = os.path.join(
            self._root_dir, climate_model, variable, region, scenario
        )
        if ensemble_member is None:
            out_fname = "{}_{}_{}_{}.nc".format(
                climate_model, variable, region, scenario
            )
        else:
            out_fname = "{}_{}_{}_{}_{}.nc".format(
                climate_model, variable, region, scenario, ensemble_member
            )

        return self._get_disk_filename(os.path.join(out_dir, out_fname))

    def save_condensed_file(self, scmrun):
        """
        Save results which have multiple ensemble members

        Parameters
        ----------
        scmrun : :obj:`scmdata.ScmRun`
            Results to save in the database

        Raises
        ------
        AssertionError
            ``ensemble_member`` is not included in ``scmrun``'s metadata
        """
        climate_model = scmrun.get_unique_meta("climate_model", no_duplicates=True)
        variable = scmrun.get_unique_meta("variable", no_duplicates=True)
        region = scmrun.get_unique_meta("region", no_duplicates=True)
        scenario = scmrun.get_unique_meta("scenario", no_duplicates=True)

        if "ensemble_member" not in scmrun.meta:
            raise AssertionError("`scmrun` must contain ensemble_member metadata")

        out_file = self.get_out_filepath(
            climate_model, variable, region, scenario, ensemble_member=None
        )
        ensure_dir_exists(out_file)
        scmrun.to_nc(out_file, dimensions=("ensemble_member",))

    def _save_to_database_single_file(self, scmrun):
        climate_model = scmrun.get_unique_meta("climate_model", no_duplicates=True)
        variable = scmrun.get_unique_meta("variable", no_duplicates=True)
        region = scmrun.get_unique_meta("region", no_duplicates=True)
        scenario = scmrun.get_unique_meta("scenario", no_duplicates=True)
        ensemble_member = scmrun.get_unique_meta("ensemble_member", no_duplicates=True)
        out_file = self.get_out_filepath(
            climate_model, variable, region, scenario, ensemble_member
        )

        ensure_dir_exists(out_file)

        scmrun.to_nc(out_file)

    def load_data(self, climate_model, variable, region, scenario):
        """
        Load data from the database

        Parameters
        ----------
        climate_model : str
            Climate model data to load

        variable : str
            Variable to load

        region : str
            Region to load

        scenario : str
            Scenario to load

        Returns
        -------
        :obj: `scmdata.ScmRun`
            Loaded data
        """
        load_path = os.path.join(
            self._root_dir, climate_model, variable, region, scenario
        )
        glob_to_use = self._get_disk_filename(os.path.join(load_path, "**", "*.nc"))
        load_files = glob.glob(glob_to_use, recursive=True)

        return run_append(
            [
                ScmRun.from_nc(f)
                for f in tqdman.tqdm(load_files, desc="Loading files", leave=False)
            ]
        )

    def save_model_reported(self, res, key="all"):
        """
        Save model reported data into the database

        Parameters
        ----------
        res : :obj:`pd.DataFrame`
            Model reported results to save. Should be the same format as the
            result of
            :meth:`pyrcmip.assessed_ranges.AssessedRanges.calculate_metric_from_results`.

        key : str
            Identifier to use in the filename

        Raises
        ------
        AssertionError
            The columns of res are not as expected (i.e.
            ``{"value", "ensemble_member", "RCMIP name", "unit", "climate_model"}``)
            or more than one climate model is included in ``res``.
        """
        expected_columns = {
            "value",
            "ensemble_member",
            "RCMIP name",
            "unit",
            "climate_model",
        }
        correct_columns = set(res.columns) == expected_columns
        if not correct_columns:
            raise AssertionError(
                "Input columns: {}. Expected columns: {}.".format(
                    set(res.columns), expected_columns
                )
            )

        climate_model = res["climate_model"].unique().tolist()
        if len(climate_model) != 1:
            raise AssertionError(
                "More than one climate model: {}".format(climate_model)
            )

        climate_model = climate_model[0]

        outfile = self._get_disk_filename(
            os.path.join(
                self._root_dir,
                climate_model,
                "model_reported_metrics_{}.csv".format(key),
            )
        )
        ensure_dir_exists(outfile)
        res.to_csv(outfile, index=False)

    def load_model_reported(self):
        """
        Load all model reported results

        Returns
        -------
        :obj:`pd.DataFrame`
            All model reported results
        """
        glob_path = self._get_disk_filename(
            os.path.join(self._root_dir, "**", "model_reported_metrics*.csv")
        )
        to_load = glob.glob(glob_path, recursive=True)

        return pd.concat([pd.read_csv(f) for f in to_load])

    def save_summary_table(self, res, file_id):
        """
        Save summary table

        Parameters
        ----------
        res : :obj:`pd.DataFrame`
            Summary table to save

        file_id : str
            Identifier to use in the filename

        Raises
        ------
        AssertionError
            Columns of ``res`` are not as expected (i.e. not equal to
            ``{"assessed_range_label", "assessed_range_value", "climate_model", "climate_model_value", "metric", "percentage_difference", "unit"}``)
        """
        expected_columns = {
            "assessed_range_label",
            "assessed_range_value",
            "climate_model",
            "climate_model_value",
            "RCMIP name",
            "percentage_difference",
            "unit",
        }
        if set(res.columns) != expected_columns:
            raise AssertionError(
                "Input columns: {}. Expected columns: {}.".format(
                    set(res.columns), expected_columns
                )
            )

        outfile = self._get_disk_filename(
            os.path.join(
                self._root_dir,
                "climate_model_assessed_ranges_summary_table_{}.csv".format(file_id),
            )
        )
        ensure_dir_exists(outfile)
        res.to_csv(outfile, index=False)

    def load_summary_tables(self):
        """
        Load all summary tables

        Returns
        -------
        :obj:`pd.DataFrame`
            All summary tables
        """
        load_path = os.path.join(
            self._root_dir,
            "climate_model_assessed_ranges_summary_table_*.csv",
        )
        load_files = glob.glob(load_path)

        return pd.concat([pd.read_csv(f) for f in load_files])
